
Unit Spacer;

Interface

Uses Dos,Graph,Objects; { Some units, we'll need 'em }

Const Pee=3.1415926535897932385; { Pi }

Type PPoint=^Point;
     PSLine=^SLine;

     Point=
     Object
      Id:Integer; { Point Id }
      X,Y,Z:Real; { Point 3D coords }
      D2X,D2Y:Integer; { Point 2D coords }
      Constructor Init(N:Word;NX,NY,NZ:Real); { Initialization }
     End;

     SLine=
     Object
      Ver1,Ver2:PPoint; { SLine points }
      Constructor Init(P,Q:PPoint); { Initialization }
     End;

Type Space3D=
     Object
      Vers,Lins:PCollection; { Points and SLines }
      WX,WY,WW,WH:Integer; { View window }
      XAng,YAng,ZAng:Real; { Angles }
      Sca,Mul:Real; { Scale and Step Multiplier }
      VerOn,LinOn:Boolean; { Points/Lines in 3D }
      PtCol,LnCol:Byte; { Colors }
      StepCn:Real; { Step counter }
      Constructor Init(NX,NY,NW,NH:Integer;S:String); { Initialization }
      Procedure SetView(NX,NY,NW,NH:Integer); { Set view window }
      Procedure Clear; { Clear object }
      Procedure Load(FName:String); { Load object from disk }
      Procedure SetScale(NSca:Real); { Scale setup }
      Procedure SetAngle(XA,YA,ZA:Real); { Angles setup }
      Procedure AutoStep; { Auto angles step }
      Procedure GetView2D(P:PPoint); { 3D to 2D conversion }
      Procedure Draw; { Display }
       Function GetPoint(N:Integer):PPoint; { Get point by Id }
     End;

Procedure WaitRetrace;

Implementation

{ Point Methods }

Constructor Point.Init(N:Word;NX,NY,NZ:Real);
Begin
 Id:=N;
 X:=NX; Y:=NY; Z:=NZ;
End;

{ SLine Methods }

Constructor SLine.Init(P,Q:PPoint);
Begin
 Ver1:=P;
 Ver2:=Q;
End;

{ Space3D Methods }

Constructor Space3D.Init(NX,NY,NW,NH:Integer;S:String);
Begin
 New(Vers,Init(1,1)); New(Lins,Init(1,1));
 XAng:=0; YAng:=0; ZAng:=0;
 Sca:=1; Mul:=0.05;
 VerOn:=False; LinOn:=True;
 PtCol:=White; LnCol:=White;
 StepCn:=0;
 SetView(NX,NY,NW,NH);
 If S<>'' Then Load(S);
End;

Procedure Space3D.SetView(NX,NY,NW,NH:Integer);
Begin
 WX:=NX; WY:=NY; WW:=NW; WH:=NH;
End;

Procedure Space3D.Clear;
Var A:Integer;
    P:PPoint;
    Q:PSLine;
Begin
 For A:=1 To Vers^.Count Do
 Begin
  P:=Vers^.At(A-1);
  Dispose(P);
 End;
 For A:=1 To Lins^.Count Do
 Begin
  Q:=Lins^.At(A-1);
  Dispose(Q);
 End;
 Vers^.DeleteAll;
 Lins^.DeleteAll;
End;

Procedure Space3D.Load(FName:String);
Var A,B,C:Integer;
    DsFile:Text;
    R,S:String;
    P,V1,V2:PPoint;
    Q:PSLine;
Begin
 If FName<>'' Then
 Begin
  S:='';
  For A:=1 To Length(FName) Do
  If (Length(S)<8) And (FName[A] In ['0'..'9','a'..'z','A'..'Z'])
  Then S:=S+FName[A];
  S:=S+'.SB';
  If (S<>'') And (FSearch(S,'')<>'') Then
  Begin
   Clear;
   Assign(DsFile,S);
   ReSet(DsFile);
   While Not EoF(DsFile) Do
   Begin
    ReadLn(DsFile,S);
    If Copy(S,1,5)='Point' Then
    Begin
     New(P,Init(0,0,0,0));
     Val(Copy(S,11,3),P^.Id,A);
     Val(Copy(S,17,7),P^.X,C);
     Val(Copy(S,27,7),P^.Y,C);
     Val(Copy(S,37,7),P^.Z,C);
     Vers^.Insert(P);
    End;
    If Copy(S,1,4)='Line' Then
    Begin
     Val(Copy(S,12,3),A,C);
     Val(Copy(S,21,3),B,C);
     V1:=GetPoint(A); V2:=GetPoint(B);
     New(Q,Init(V1,V2));
     Lins^.Insert(Q);
    End;
   End;
   Close(DsFile);
  End;
 End;
End;

Procedure Space3D.SetScale(NSca:Real);
Begin
 Sca:=NSca;
End;

Procedure Space3D.SetAngle(XA,YA,ZA:Real);
Begin
 XAng:=XA; YAng:=YA; ZAng:=ZA;
End;

Procedure Space3D.AutoStep;
Var SA,CA:Real;
Begin
 StepCn:=StepCn+0.01;
 SA:=Sin(StepCn)*Mul;
 CA:=Cos(StepCn)*Mul;
 SetAngle(XAng+SA,YAng+CA,ZAng+(SA+CA)*0.5);
End;

Procedure Space3D.GetView2D(P:PPoint);
Var Y3X,Z3X,X3Y,Z3Y,X3Z,Y3Z:Real;
    CX,SX,CY,SY,CZ,SZ:Real;
Begin
 CX:=Cos(XAng); SX:=Sin(XAng);
 CY:=Cos(YAng); SY:=Sin(YAng);
 CZ:=Cos(ZAng); SZ:=Sin(ZAng);
 Y3X:=P^.Y*CX+P^.Z*SX; Z3X:=P^.Z*CX-P^.Y*SX;
 X3Y:=P^.X*CY+Z3X*SY; Z3Y:=Z3X*CY-P^.X*SY;
 X3Z:=X3Y*CZ+Y3X*SZ; Y3Z:=Y3X*CZ-X3Y*SZ;
 P^.D2X:=Round(WX+WW/2+X3Z*Sca); P^.D2Y:=Round(WY+WH/2+Y3Z*Sca);
End;

Procedure Space3D.Draw;
Var A,B:Integer;
    P:PPoint;
    Q:PSLine;
    X1,Y1,X2,Y2:Integer;
Begin
 For A:=1 To Vers^.Count Do
 Begin
  P:=Vers^.At(A-1);
  GetView2D(P);
  If VerOn Then
  Begin
   X1:=P^.D2X; Y1:=P^.D2Y;
   PutPixel(X1,Y1,PtCol);
  End;
 End;
 If LinOn Then For A:=1 To Lins^.Count Do
 Begin
  Q:=Lins^.At(A-1);
  X1:=Q^.Ver1^.D2X; Y1:=Q^.Ver1^.D2Y;
  X2:=Q^.Ver2^.D2X; Y2:=Q^.Ver2^.D2Y;
  SetColor(LnCol);
  Line(X1,Y1,X2,Y2);
 End;
End;

Function Space3D.GetPoint(N:Integer):PPoint;
Var A:Integer;
    P:PPoint;
    Ok:Boolean;
Begin
 A:=1;
 Ok:=False;
 While (A<=Vers^.Count) And (Not Ok) Do
 Begin
  P:=Vers^.At(A-1);
  If P^.Id=N Then Ok:=True;
  Inc(A);
 End;
 GetPoint:=P;
End;

Procedure WaitRetrace; Assembler;
Asm
 Mov DX,3DAh
 @1:
  In AL,DX
 And AL,08h
 Jnz @1
 @2:
  In AL,DX
 And AL,08h
  Jz @2
End;

End.